import pathlib
from typing import Any
from unittest import TestCase
from unittest.mock import Mock, call, patch
from uuid import uuid4

from box import Box

from pycarlo.common.files import BytesFileReader, JsonFileReader
from pycarlo.features.dbt import DbtImporter
from pycarlo.features.dbt.queries import (
    GET_DBT_UPLOAD_URL,
    SEND_DBT_ARTIFACTS_EVENT,
)
from pycarlo.features.pii import PiiFilterer
from pycarlo.features.user import Resource


class DbtImportServiceTest(TestCase):
    manifest_path = f"{pathlib.Path(__file__).parent}/sample_manifest.json"
    run_results_path = f"{pathlib.Path(__file__).parent}/sample_run_results.json"
    logs_path = f"{pathlib.Path(__file__).parent}/sample_logs.txt"

    def setUp(self) -> None:
        self._mock_pii_service = Mock()
        self._mock_pii_service.get_pii_filters_config.return_value = None

    @patch("pycarlo.features.dbt.dbt_importer.http")
    def test_import_run(self, mock_http: Mock):
        # given
        resource = Resource(id=uuid4(), name="Snowflake", type="snowflake")
        mock_user_service = Mock()
        mock_user_service.get_resource.return_value = resource

        def mock_client_responses(**kwargs: Any):
            query = kwargs["query"]
            if query == GET_DBT_UPLOAD_URL:
                return Box({"get_dbt_upload_url": f"https://{kwargs['variables']['fileName']}"})

        mock_client = Mock(side_effect=mock_client_responses)

        importer = DbtImporter(
            mc_client=mock_client,
            user_service=mock_user_service,
            pii_service=self._mock_pii_service,
        )

        # when
        importer.import_run(
            manifest_path=self.manifest_path,
            run_results_path=self.run_results_path,
            logs_path=self.logs_path,
            resource_id=resource.id,
        )

        # verify expected call to user service
        mock_user_service.get_resource.assert_called_once_with(resource.id)

        # verify expected calls to upload artifacts to S3
        self.assertEqual(3, mock_http.upload.call_count)
        mock_http.upload.assert_has_calls(
            [
                call(
                    method="put",
                    url="https://sample_manifest.json",
                    content=JsonFileReader(self.manifest_path).read(),
                ),
                call(
                    method="put",
                    url="https://sample_run_results.json",
                    content=JsonFileReader(self.run_results_path).read(),
                ),
                call(
                    method="put",
                    url="https://sample_logs.txt",
                    content=BytesFileReader(self.logs_path).read(),
                ),
            ]
        )

        # verify expected MC client calls
        self.assertEqual(4, mock_client.call_count)
        mock_client.assert_has_calls(
            [
                call(
                    query=GET_DBT_UPLOAD_URL,
                    variables=dict(
                        projectName="default-project",
                        invocationId="3b44f6e7-0a4a-4c81-8859-468b2d15075e",
                        fileName="sample_manifest.json",
                    ),
                    additional_headers={
                        "x-mcd-telemetry-reason": "service",
                        "x-mcd-telemetry-service": "dbt_importer",
                    },
                ),
                call(
                    query=GET_DBT_UPLOAD_URL,
                    variables=dict(
                        projectName="default-project",
                        invocationId="3b44f6e7-0a4a-4c81-8859-468b2d15075e",
                        fileName="sample_run_results.json",
                    ),
                    additional_headers={
                        "x-mcd-telemetry-reason": "service",
                        "x-mcd-telemetry-service": "dbt_importer",
                    },
                ),
                call(
                    query=GET_DBT_UPLOAD_URL,
                    variables=dict(
                        projectName="default-project",
                        invocationId="3b44f6e7-0a4a-4c81-8859-468b2d15075e",
                        fileName="sample_logs.txt",
                    ),
                    additional_headers={
                        "x-mcd-telemetry-reason": "service",
                        "x-mcd-telemetry-service": "dbt_importer",
                    },
                ),
                call(
                    query=SEND_DBT_ARTIFACTS_EVENT,
                    variables=dict(
                        projectName="default-project",
                        jobName="default-job",
                        invocationId="3b44f6e7-0a4a-4c81-8859-468b2d15075e",
                        artifacts=dict(
                            manifest="sample_manifest.json",
                            runResults="sample_run_results.json",
                            logs="sample_logs.txt",
                        ),
                        resourceId=str(resource.id),
                    ),
                    additional_headers={
                        "x-mcd-telemetry-reason": "service",
                        "x-mcd-telemetry-service": "dbt_importer",
                    },
                ),
            ]
        )

    @patch("pycarlo.features.dbt.dbt_importer.http")
    @patch("pycarlo.features.dbt.DbtImporter._init_pii_filterer")
    def test_import_run_filtered(self, mock_init_filterer: Mock, mock_http: Mock):
        resource = Resource(id=uuid4(), name="Snowflake", type="snowflake")
        mock_user_service = Mock()
        mock_user_service.get_resource.return_value = resource

        mock_init_filterer.return_value = PiiFilterer(
            filters_config={"active": [{"name": "thread-id", "pattern": r"Thread-\d{2}"}]},
            include_metrics=False,
        )

        def mock_client_responses(**kwargs: Any):
            query = kwargs["query"]
            if query == GET_DBT_UPLOAD_URL:
                return Box({"get_dbt_upload_url": f"https://{kwargs['variables']['fileName']}"})

        mock_client = Mock(side_effect=mock_client_responses)

        importer = DbtImporter(
            mc_client=mock_client,
            user_service=mock_user_service,
            pii_service=self._mock_pii_service,
        )

        # when
        importer.import_run(
            manifest_path=self.manifest_path,
            run_results_path=self.run_results_path,
            logs_path=self.logs_path,
            resource_id=resource.id,
        )

        expected_run_results = JsonFileReader(self.run_results_path).read()
        for r in expected_run_results["results"]:
            r["thread_id"] = "<filtered:thread-id>"

        # verify call was properly filtered, we know there are thread ids in sample run results
        mock_http.upload.assert_has_calls(
            [
                call(
                    method="put",
                    url="https://sample_run_results.json",
                    content=expected_run_results,
                ),
            ]
        )



import logging
import re
import time
from dataclasses import dataclass
from re import Pattern
from typing import Any, Callable, Dict, List, Optional, Union

from dataclasses_json import dataclass_json

_logger = logging.getLogger(__name__)


@dataclass_json
@dataclass
class PiiActiveFilter:
    name: str
    pattern: str


@dataclass_json
@dataclass
class PiiActiveFiltersConfig:
    active: List[PiiActiveFilter]
    fail_closed: bool = True

    @staticmethod
    def is_valid_config(cfg: Optional[Dict]) -> bool:
        return bool(cfg and cfg.get("active", []))


@dataclass_json
@dataclass
class PiiFilterMetrics:
    replacements: int
    time_taken_ms: float


@dataclass
class PiiCompiledFilter:
    name: str
    compiled_expression: Pattern
    replacement: str


class PiiFilterer:
    _TOTAL_KEY = "_total"
    _REPLACEMENT_STRING = "<filtered:{}>"

    def __init__(self, filters_config: Optional[Dict], include_metrics: bool = True):
        self._config = (
            PiiActiveFiltersConfig.from_dict(filters_config)  # type: ignore
            if PiiActiveFiltersConfig.is_valid_config(filters_config)
            else None
        )

        if self._config:
            self._filters = [
                PiiCompiledFilter(
                    name=f.name,
                    compiled_expression=re.compile(f.pattern),
                    replacement=self._REPLACEMENT_STRING.format(f.name),
                )
                for f in self._config.active
            ]
        self._include_metrics = include_metrics

    @staticmethod
    def _metrics_to_final(metrics: Dict, elapsed_time_ns: int) -> Dict:
        final_metrics = {
            k: PiiFilterMetrics(replacements=v, time_taken_ms=t / 1000000).to_dict()  # type: ignore
            for k, (v, t) in metrics.items()
        }
        total_replacements = sum(m["replacements"] for m in final_metrics.values())

        final_metrics[PiiFilterer._TOTAL_KEY] = PiiFilterMetrics(
            replacements=total_replacements, time_taken_ms=elapsed_time_ns / 1000000
        ).to_dict()  # type: ignore

        return final_metrics

    def filter_content(self, content: Union[bytes, str, Dict]) -> Union[bytes, str, Dict]:
        """
        Utility method to filter one of bytes, str or Dict. Calls internally one of:
        - filter_data
        - filter_str
        - filter_message
        """
        if isinstance(content, dict):
            return self.filter_message(content)
        elif isinstance(content, str):
            return self.filter_str(content)
        else:
            return self.filter_data(content)

    def filter_message(self, msg: Union[bytes, str, Dict]) -> Union[bytes, str, Dict]:
        """
        Filters a dictionary or a list. If the object is a dictionary it includes metrics in a
        pii_metrics attribute in the result dictionary (only if the filterer was created with
        include_metrics=True which is the default value).
        """
        if not self._config:
            return msg

        start_time = time.time_ns()

        include_metrics = self._include_metrics and isinstance(msg, Dict)
        metrics = {} if include_metrics else None

        result = self._do_filter(msg, lambda o: self._filter_object(o, metrics=metrics))

        if include_metrics:
            assert metrics is not None
            elapsed_time = time.time_ns() - start_time
            result["pii_metrics"] = self._metrics_to_final(metrics, elapsed_time_ns=elapsed_time)

        return result

    def _do_filter(self, data: Any, filter_function: Callable[[Any], Any]) -> Any:
        if not self._config:
            return data
        try:
            result = filter_function(data)
        except Exception as exc:
            if self._config.fail_closed:
                raise
            _logger.exception(
                f"Failed to evaluate PII filters: {exc}, ignoring because fail_closed=False"
            )
            result = data
        return result

    def filter_str(self, msg: str) -> str:
        """
        Filters a string, please note metrics are not included as we don't have a way to include
        metrics in the resulting string. We might return metrics in the future if we decide to
        use them where we call this (dbtv2 collection)
        """
        return self._do_filter(msg, lambda o: self._filter_text(o))

    def filter_data(self, msg: bytes, encoding: str = "utf8") -> bytes:
        """
        Filters a bytes array using the given encoding, please note metrics are not included as
        we don't have a way to include metrics in the resulting bytes array. We might return metrics
        in the future if we decide to use them where we call this (dbtv2 collection)
        """
        return self._do_filter(msg, lambda o: self._filter_bytes(o, encoding=encoding))

    def _filter_bytes(self, data: bytes, encoding: str) -> bytes:
        text = data.decode(encoding)
        filtered_text = self._filter_text(text)
        return filtered_text.encode(encoding)

    def _filter_object(self, o: Any, metrics: Optional[Dict] = None) -> Any:
        if isinstance(o, Dict):
            return {
                self._filter_text(k, metrics=metrics): self._filter_object(v, metrics=metrics)
                for k, v in o.items()
            }
        elif isinstance(o, List):
            return [self._filter_object(e, metrics=metrics) for e in o]
        elif isinstance(o, tuple):
            return tuple(self._filter_object(e, metrics=metrics) for e in o)
        elif isinstance(o, str):
            return self._filter_text(o, metrics=metrics)
        else:
            return o

    def _filter_text(self, text: str, metrics: Optional[Dict] = None) -> str:
        for f in self._filters:
            filter_name = f.name
            start_time = time.time_ns()
            text, updated_count = f.compiled_expression.subn(f.replacement, text)
            elapsed_time = time.time_ns() - start_time

            if metrics is not None:
                prev_count, prev_time = metrics.get(filter_name, (0, 0))
                metrics[filter_name] = (prev_count + updated_count, prev_time + elapsed_time)

        return text
